@@ -1,6 +1,7 @@
 #!/usr/bin/python
 """
-This program moves local files to and from Amazon's Simple Storage System.
+This program moves local files to and from Amazon's Simple Storage
+System.
 
 Capability:
   - Test if file exists in S3 bucket
@@ -19,11 +20,12 @@ S3 files also have metadata in addition to their content.
 metadata is a python dict i.e. a set of key, value pairs.
 
 IMPLEMENTATION NOTES:
-The implementation is based on tinys3.  I've kept the connection and request 
-objects but changed their API so it is not compatible with tinys3 at all.
-I do use tinys3's auth module with a change to the signing logic.
+The implementation is based on tinys3.  I've kept the connection 
+and request objects but changed their API so it is not compatible 
+with tinys3 at all.  I do use tinys3's util module unchanged, and 
+auth module with a change to the signing logic.
 
-Paul Wexler, Apr 17, 2014
+Paul Wexler, Jun 3, 2014
 
 Copyright (C) 2014, Prometheus Research, LLC
 """
@@ -46,7 +48,7 @@ MB = 1024 * 1024
 GB = 1024 * MB
 TB = 1024 * GB
 
-version = '0.0.2'
+version = '0.1.0'
 
 class S3Facts:
     amz_metadata_prefix = 'x-amz-meta-'
@@ -57,16 +59,17 @@ class S3Facts:
     multipart_threshhold = 100 * MB
     
 class S3Name:
-    """An S3 file name consists of a bucket and a key.  This pair uniquely 
-    identifies the file within S3.  
+    """An S3 file name consists of a bucket and a key.  This pair 
+    uniquely identifies the file within S3.  
     
-    The S3Connection class methods take a remote_name argument which can be 
-    either a string which is the key, or an instance of the S3Name class.  
-    When no bucket is given the default_bucket established when the connection 
-    is instantiated is used.
+    The S3Connection class methods take a remote_name argument 
+    which can be either a string which is the key, or an instance 
+    of the S3Name class.  When no bucket is given the 
+    default_bucket established when the connection is instantiated 
+    is used.
     
-    In other words, the S3Name class provides a means of using a bucket other 
-    than the default_bucket.
+    In other words, the S3Name class provides a means of using a 
+    bucket other than the default_bucket.
     """
     def __init__(self, key, bucket=None):
         self.key = key
@@ -277,8 +280,8 @@ class ExistsS3Request(S3Request):
         
 class ReadS3Request(S3Request): 
     """
-    send() returns S3Response instance which includes a metadata attribute
-    containing the metadata of the bucket/key.
+    send() returns S3Response instance which includes a metadata 
+    attribute containing the metadata of the bucket/key.
     """
     def __init__(self,  conn, bucket, key, local_file):
         super(ReadS3Request, self).__init__(conn)
@@ -523,9 +526,12 @@ class S3Connection(object):
         Params:
         access_key_id   - AWS access key (username)
         secret_access_key - AWS secret access key (password)
-        default_bucket  - Sets the default bucket.  Default is None
-        tls             - do/not use secure connection.  Default is True.
-        endpoint        - Sets the s3 endpoint.  Default is s3.amazonaws.com
+        default_bucket  - Sets the default bucket.  
+                          Default is None
+        tls             - do/not use secure connection.  
+                          Default is True.
+        endpoint        - Sets the s3 endpoint.  
+                          Default is s3.amazonaws.com
         """
         self.auth = S3Auth(access_key_id, secret_access_key)
         self.default_bucket = default_bucket
@@ -582,11 +588,11 @@ class S3Connection(object):
         """
         Verifies that we have a bucket for a request
         Params:
-        bucket -    The name of the bucket we're trying to use, None if we want 
-                    to use the default bucket.
+        bucket -    The name of the bucket we're trying to use, 
+                    None if we want to use the default bucket.
         Returns:    The bucket to use for the request
-        Raises:
-        ValueError if no bucket was provided AND no default bucket was defined.
+        Raises:     ValueError if no bucket was provided AND no 
+                    default bucket was defined.
         """
         b = bucket or self.default_bucket
         if not b:
@@ -606,141 +612,168 @@ class S3Connection(object):
             raise Exception, 'invalid remote_name: %s' % remote_name
         return self._get_bucket(bucket), key
 
-class RemoteStoreError(Exception):
-    pass 
+class StorageError(Exception):
+    def __init__(self, msg, exception):
+        self.msg = msg
+        self.exception = exception
     
-class RemoteStore(object):
-    """RemoteStore
-
+    def __str__(self):
+        return '%s: %s' % (self.msg, self.exception)
+    
+class Storage(object):
+    """
 API to remote storage
 =====================
 
-**ok = store.copy(remote_src, remote_dst, headers={})**
-    Copy from remote store to remote store.  Used for renaming remote files or 
-    updating their metadata.
-**ok = store.delete(remote_name)**
-    Delete file from remote store.
-**exists, metadata = store.exists(remote_name)**
-    Test if file exists in remote store and retrieve its metadata if it does.
-**ok, metadata = store.read(remote_name, local_name)**
-    Download a file from remote store and retrieve its metadata.
-**ok = store.update_metadata(remote_name, headers)**
-    Update a remote file's metadata.
-**ok = store.write(local_name, remote_name, headers={})**
-    Upload a file to remote store (and possibly set its metadata).
-
-When a request is sent, the requests module may raise an exception or may
-return status.  The status could indicate success or failure.  So there 
-are 2 kinds of failure.  
-
-All methods except **exists** catch connection exceptions, log either type of 
-failure, and return True for success, and False for failure.
-
-The **exists** method does not catch connection exceptions and may also raise 
-RemoteStoreError.  It returns True or False as the remote_name exists in
-remote storage or not.
-
-**copy**, **update_metadata**, and **write** accept a headers argument which 
-is used to provide additional headers to the request.
-
-**exists** returns a tuple whose first element indicates if the file exists or 
-not, and whose second element is the file's metadata (when the file exists).
-
-**read** returns a tuple whose first element indicates success or failure and 
-whose second element is the file's metadata (on success).
-
 S3 Filenames
 ------------
-An S3 file name consists of a bucket and a key.  This pair uniquely 
-identifies the file within S3.  
 
-The S3Connection class methods take a remote_name argument which can be 
-either a string which is the key, or an instance of the S3Name class.  
-When no bucket is given the default_bucket established when the connection 
-is instantiated is used.
+An S3 file name consists of a bucket and a key.  This pair of 
+strings uniquely identifies the file within S3.  
 
-In other words, the S3Name class provides a means of using a bucket other 
-than the default_bucket.
+The S3Name class is instantiated with a key and a bucket; the key 
+is required and the bucket defaults to None.
+
+The RemoteStore class methods take a *remote_name* argument which 
+can be either a string which is the key, or an instance of the 
+S3Name class.  When no bucket is given (or the bucket is None) then 
+the default_bucket established when the connection is instantiated 
+is used.  If no bucket is given (or the bucket is None) and there 
+is no default bucket then a ValueError is raised.
+
+In other words, the S3Name class provides a means of using a bucket 
+other than the default_bucket.
 
 Headers and Metadata
 --------------------
 
-Additional http headers may be sent with the methods which write data.  These
-control various aspects of how the file may be handled.  S3 supports a variety
-of headers.  Those headers whose key begins with the special preifx:
-'x-amz-meta-' are considered to be metadata headers and are used to set 
-the metadata attributes of the file.
+Additional http headers may be sent using the methods which write 
+data.  These methods accept an optional *headers* argument which 
+is a python dict.  The headers control various aspects of how the 
+file may be handled.  S3 supports a variety of headers.  These are 
+not discussed here.  See Amazon's S3 documentation for more info
+on S3 headers.  Those headers whose key begins with the special 
+preifx: 'x-amz-meta-' are considered to be metadata headers and are
+used to set the metadata attributes of the file.
+
+The methods which read files also return metadata which consists of 
+only those response headers which begin with 'x-amz-meta-'.
+
+Storage Methods
+---------------
+
+The arguments *remote_source*, *remote_destination*, and 
+*remote_name* may be either a string, or an S3Name instance.
+
+*local_name* is a string and is the name of the file on the local 
+system.  This string is passed directly to open().
+
+*headers* is a python dict used to encode additional request 
+headers.
+
+All methods return on success or raise StorageError on failure.
+
+**storage.copy(remote_source, remote_destination, headers={})**
+    Copy *remote_source* to *remote_destination*.  The destination 
+    metadata is copied from *headers* when it contains metadata; 
+    otherwise it is copied from the source metadata.
+**storage.delete(remote_name)**
+    Delete *remote_name* from storage.
+**exists, metadata = storage.exists(remote_name)**
+    Test if *remote_name* exists in storage, retrieve its metadata 
+    if it does.
+    exists - boolean, metadata - dict.
+**metadata = storage.read(remote_name, local_name)**
+    Download *remote_name* from storage, save it locally as 
+    *local_name* and retrieve its metadata.
+    metadata - dict.
+**storage.update_metadata(remote_name, headers)**
+    Update the metadata associated with *remote_name* with the 
+    metadata headers in *headers*.
+**storage.write(local_name, remote_name, headers={})**
+    Upload *local_name* to storage as *remote_name*, and set its 
+    metadata if any metadata headers are in *headers*.
 """
     def __init__(self, connection, **kwargs):
         self.connection = connection
 
     def copy(self, remote_src, remote_dst, headers={}):
         """
-        Copies remote_src to remote_dst.  headers may be used to change the 
-        properties of remote_dst including its metadata.
+        Copies remote_src to remote_dst.  headers may be used to 
+        change the properties of remote_dst including its metadata.
         
-        If any metadata headers are included in headers then they become the 
-        metadata of remote_dst; otherwise the metadata of remote_src is copied.
+        If any metadata headers are included in headers then they 
+        become the metadata of remote_dst; otherwise the metadata 
+        of remote_src is copied.
         
         A metadata header is any header that starts with 
         `S3Facts.amz_metadata_prefix`
         
-        Returns True/False => success/failure.
-        In the event of failure, an error message is logged.
+        Returns on success, raises StorageError on failure.
         """
         try:
             response = self.connection.copy(
                     remote_src, 
                     remote_dst, 
                     headers=headers)
-            if not response.ok:
-                logging.error(response.error)
-            return response.ok
-
+            if response.ok:
+                return
+            else:
+                raise StorageError('copy', response.error)
         except Exception, e:
-            logging.exception(e)
-            return False
+            raise StorageError('copy ', e)
 
     def delete(self, remote_name):
         """
-        Deletes remote_name from the remote storage
-        Returns True/False => success/failure.
-        In the event of failure, an error message is logged.
-        """
+        Deletes remote_name from storage.
+        Returns on success, raises StorageError on failure.
+       """
         try:
             response = self.connection.delete(remote_name)
-            if not response.ok:
-                logging.error(response.error)
-            return response.ok
-
+            if response.ok:
+                return
+            else:
+                raise StorageError('delete', response.error)
         except Exception, e:
-            logging.exception(e)
-            return False
+            raise StorageError('delete ', e)
 
     def exists(self, remote_name):
         """
         Tests if remote_name is stored in remote storage.
         Returns (exists, metadata)
             exists:  True if remote_name exists, False otherwise.
-            metadata: the file's metadata (when exists == True)
-        This method does not catch connection exceptions and may also throw 
-        a RemoteStoreError exception.
+            metadata: a dict.  The file's metadata (when exists 
+            == True)
+        raises StorageError on failure.
         """
-        response = self.connection.exists(remote_name)
-        if response.ok:
-            return not response.error, response.metadata
-        else:
-            raise RemoteStoreError, 'exists: %s' % response.error
-
+        try:
+            response = self.connection.exists(remote_name)
+            if response.ok:
+                return not response.error, response.metadata
+            else:
+                raise RemoteStoreError, 'exists: %s' % response.error
+        
+        # Earlier, when the file does not exist, requests returned a 404 
+        # with a non-empty reason.  Something must have changed!  
+        #
+        # Now a ChunkedEncodingError is raised with 
+        # "IncompleteRead(0 bytes read)"
+        #
+        except requests.models.ChunkedEncodingError, e:
+            if str(e) == 'IncompleteRead(0 bytes read)':
+                return False, {}
+            else:
+                raise StorageError('exists', e)
+        except Exception, e:
+            raise StorageError('exists ', e)
+    
     def read(self, remote_name, local_name):
         """
-        Reads (downloads) `remote_name` from remote storage and saves it 
+        Reads (downloads) `remote_name` from storage and saves it 
         in the local file system as `local_name`
         
-        Returns (ok, metadata)
-            ok:  True/False => success/failure.
-                 In the event of failure, an error message is logged.
-            metadata: the file's metadata (when ok == True)
+        Returns metadata (a dict) on success, raises StorageError 
+        on failure.
         """
         try:
             timestamp = time.strftime('%Y%m%d%H%M%S', time.localtime())
@@ -749,47 +782,43 @@ the metadata attributes of the file.
                 response = self.connection.read(remote_name, fo)
             if response.ok:
                 os.rename(tmp_filename, local_name)
+                return response.metadata
             else:
-                logging.error(response.error)
-            return response.ok, response.metadata
+                raise StorageError('read', response.error)
         except Exception, e:
-            logging.exception(e)
-            return False, {}
+            raise StorageError('read ', e)
 
     def update_metadata(self, remote_name, headers):
         """
-        Updates the properties of remote_name by including headers in the 
-        request.
+        Updates the properties of remote_name by including headers
+        in the request.
 
-        Headers which begin with `S3Facts.amz_metadata_prefix` are used to
-        set metadata.
+        Headers which begin with `S3Facts.amz_metadata_prefix` are
+        used to set metadata.
         
-        Returns True/False => success/failure.
-        In the event of failure, an error message is logged.
+        Returns on success, raises StorageError on failure.
         """
         try:
             response = self.connection.update_metadata(
                     remote_name,
                     headers=headers)
-            if not response.ok:
-                logging.error(response.error)
-            return response.ok
-
+            if response.ok:
+                return
+            else:
+                raise StorageError('update_metadata', response.error)
         except Exception, e:
-            logging.exception(e)
-            return False
+            raise StorageError('update_metadata ', e)
 
     def write(self, local_name, remote_name, headers={}):
         """
-        Writes (uploads) `local_name` from the local file system to remote 
-        storage as `remote_name`. `headers` may contain additional headers for 
-        the request. 
+        Writes (uploads) `local_name` from the local file system 
+        to storage as `remote_name`. `headers` may contain 
+        additional headers for the request. 
         
-        Headers which begin with `S3Facts.amz_metadata_prefix` are used to
-        set metadata.
+        Headers which begin with `S3Facts.amz_metadata_prefix` are
+        used to set metadata.
         
-        Returns True/False => success/failure.
-        In the event of failure, an error message is logged.
+        Returns on success, raises StorageError on failure.
         """
         try:
             with open(local_name, 'r') as fi:
@@ -797,10 +826,10 @@ the metadata attributes of the file.
                         fi, 
                         remote_name,
                         headers=headers)
-            if not response.ok:
-                logging.error(response.error)
-            return response.ok
+            if response.ok:
+                return
+            else:
+                raise StorageError('write', response.error)
         except Exception, e:
-            logging.exception(e)
-            return False
+            raise StorageError('write ', e)
 
